import { EVMBlockchainIncludingTestnet } from '@crossmint/common-sdk-base';
export { EVMBlockchainIncludingTestnet as Blockchain, blockchainToChainId } from '@crossmint/common-sdk-base';
import { LocalAccount, EIP1193Provider, HttpTransport, Chain, Hex, PublicClient } from 'viem';
import { PasskeyValidatorContractVersion } from '@zerodev/passkey-validator';
import { SmartAccountClient } from 'permissionless';
import { SmartAccount } from 'permissionless/accounts';
import { EntryPoint } from 'permissionless/types/entrypoint';

type SmartWalletSDKInitParams = {
    clientApiKey: string;
};
type UserParams = {
    jwt: string;
};
type ViemAccount = {
    type: "VIEM_ACCOUNT";
    account: LocalAccount & {
        source: "custom";
    };
};
type PasskeySigner = {
    type: "PASSKEY";
    /**
     * Displayed to the user during passkey registration or signing prompts.
     * If not provided, a default name identifier within the JWT
     * that is specified in the project settings (typically `sub`) will be used.
     */
    passkeyName?: string;
};
type EOASigner = EIP1193Provider | ViemAccount;
interface WalletParams {
    signer: EOASigner | PasskeySigner;
}

declare const SUPPORTED_KERNEL_VERSIONS: readonly ["0.3.1", "0.3.0", "0.2.4"];
type SupportedKernelVersion = (typeof SUPPORTED_KERNEL_VERSIONS)[number];
declare const SUPPORTED_ENTRYPOINT_VERSIONS: readonly ["v0.6", "v0.7"];
type SupportedEntryPointVersion = (typeof SUPPORTED_ENTRYPOINT_VERSIONS)[number];
type PasskeyValidatorSerializedData = {
    passkeyServerUrl: string;
    entryPoint: Hex;
    validatorAddress: Hex;
    pubKeyX: string;
    pubKeyY: string;
    authenticatorIdHash: Hex;
    authenticatorId: string;
};
type SmartWalletClient = SmartAccountClient<EntryPoint, HttpTransport, Chain, SmartAccount<EntryPoint>>;

type StoreSmartWalletParams = {
    type: string;
    smartContractWalletAddress: string;
    signerData: SignerData;
    sessionKeySignerAddress?: string;
    version: number;
    baseLayer: string;
    chainId: number;
    entryPointVersion: SupportedEntryPointVersion;
    kernelVersion: SupportedKernelVersion;
};
type SignerData = EOASignerData | PasskeySignerData;
interface EOASignerData {
    eoaAddress: string;
    type: "eoa";
}
type PasskeySignerData = PasskeyValidatorSerializedData & {
    passkeyName: string;
    validatorContractVersion: PasskeyValidatorContractVersion;
    domain: string;
    type: "passkeys";
};
type PasskeyDisplay = Pick<PasskeySignerData, "type" | "passkeyName" | "pubKeyX" | "pubKeyY">;
type SignerDisplay = EOASignerData | PasskeyDisplay;

declare class LoggerWrapper {
    private extraInfo;
    private logIdempotencyKey;
    constructor(className: string, extraInfo?: {}, logIdempotencyKey?: string);
    private logInput;
    private logOutput;
    private logError;
    protected logPerformance<T>(name: string, cb: () => Promise<T>): Promise<T>;
}

declare const SmartWalletErrors: {
    readonly NOT_AUTHORIZED: "smart-wallet:not-authorized";
    readonly TRANSFER: "smart-wallet:transfer.error";
    readonly CROSSMINT_SERVICE: "smart-wallet:crossmint-service.error";
    readonly ERROR_JWT_EXPIRED: "smart-wallet:not-authorized.jwt-expired";
    readonly ERROR_JWT_INVALID: "smart-wallet:not-authorized.jwt-invalid";
    readonly ERROR_JWT_DECRYPTION: "smart-wallet:not-authorized.jwt-decryption";
    readonly ERROR_JWT_IDENTIFIER: "smart-wallet:not-authorized.jwt-identifier";
    readonly ERROR_USER_WALLET_ALREADY_CREATED: "smart-wallet:user-wallet-already-created.error";
    readonly ERROR_OUT_OF_CREDITS: "smart-wallet:out-of-credits.error";
    readonly ERROR_WALLET_CONFIG: "smart-wallet:wallet-config.error";
    readonly ERROR_ADMIN_MISMATCH: "smart-wallet:wallet-config.admin-mismatch";
    readonly ERROR_PASSKEY_MISMATCH: "smart-wallet:wallet-config.passkey-mismatch";
    readonly ERROR_ADMIN_SIGNER_ALREADY_USED: "smart-wallet:wallet-config.admin-signer-already-used";
    readonly UNCATEGORIZED: "smart-wallet:uncategorized";
};
type SmartWalletErrorCode = (typeof SmartWalletErrors)[keyof typeof SmartWalletErrors];
declare class SmartWalletSDKError extends Error {
    readonly code: SmartWalletErrorCode;
    readonly details?: string;
    constructor(message: string, details?: string, code?: SmartWalletErrorCode);
}
declare class TransferError extends SmartWalletSDKError {
    constructor(message: string);
}
declare class CrossmintServiceError extends SmartWalletSDKError {
    status?: number;
    constructor(message: string, status?: number);
}
declare class AdminMismatchError extends SmartWalletSDKError {
    readonly required: SignerDisplay;
    readonly used?: SignerDisplay;
    constructor(message: string, required: SignerDisplay, used?: SignerDisplay);
}
declare class PasskeyMismatchError extends SmartWalletSDKError {
    readonly required: PasskeyDisplay;
    readonly used?: PasskeyDisplay;
    constructor(message: string, required: PasskeyDisplay, used?: PasskeyDisplay);
}
declare class NotAuthorizedError extends SmartWalletSDKError {
    constructor(message: string);
}
declare class JWTExpiredError extends NotAuthorizedError {
    readonly code: "smart-wallet:not-authorized.jwt-expired";
    /**
     * The expiry time of the JWT as an ISO 8601 timestamp.
     */
    readonly expiredAt: string;
    constructor(expiredAt: Date);
}
declare class JWTInvalidError extends NotAuthorizedError {
    readonly code: "smart-wallet:not-authorized.jwt-invalid";
    constructor();
}
declare class JWTDecryptionError extends NotAuthorizedError {
    readonly code: "smart-wallet:not-authorized.jwt-decryption";
    constructor();
}
declare class JWTIdentifierError extends NotAuthorizedError {
    readonly code: "smart-wallet:not-authorized.jwt-identifier";
    readonly identifierKey: string;
    constructor(identifierKey: string);
}
declare class UserWalletAlreadyCreatedError extends SmartWalletSDKError {
    readonly code: "smart-wallet:user-wallet-already-created.error";
    constructor(userId: string);
}
declare class OutOfCreditsError extends SmartWalletSDKError {
    constructor(message?: string);
}
declare class ConfigError extends SmartWalletSDKError {
    constructor(message: string);
}
declare class AdminAlreadyUsedError extends ConfigError {
    readonly code: "smart-wallet:wallet-config.admin-signer-already-used";
    constructor();
}
declare class NonCustodialWalletsNotEnabledError extends ConfigError {
    constructor();
}

type CrossmintAPIErrorCodes = "ERROR_JWT_INVALID" | "ERROR_JWT_DECRYPTION" | "ERROR_JWT_IDENTIFIER" | "ERROR_JWT_EXPIRED" | "ERROR_USER_WALLET_ALREADY_CREATED" | "ERROR_ADMIN_SIGNER_ALREADY_USED" | "ERROR_PROJECT_NONCUSTODIAL_WALLETS_NOT_ENABLED";
declare class APIErrorService {
    private errors;
    constructor(errors?: Partial<Record<CrossmintAPIErrorCodes, (apiResponse: any) => SmartWalletSDKError>>);
    throwErrorFromResponse({ response, onServerErrorMessage, }: {
        response: Response;
        onServerErrorMessage: string;
    }): Promise<void>;
}

declare abstract class BaseCrossmintService extends LoggerWrapper {
    crossmintAPIHeaders: Record<string, string>;
    protected crossmintBaseUrl: string;
    protected apiErrorService: APIErrorService;
    private static urlMap;
    constructor(apiKey: string);
    protected fetchCrossmintAPI(endpoint: string, options: {
        body?: string;
        method: string;
    } | undefined, onServerErrorMessage: string, authToken?: string): Promise<any>;
    protected getUrlFromEnv(environment: string): string;
}

declare class CrossmintWalletService extends BaseCrossmintService {
    idempotentCreateSmartWallet(user: UserParams, input: StoreSmartWalletParams): Promise<any>;
    getSmartWalletConfig(user: UserParams, chain: EVMBlockchainIncludingTestnet): Promise<{
        kernelVersion: string;
        entryPointVersion: string;
        userId: string;
        signers: {
            signerData: SignerData;
        }[];
        smartContractWalletAddress?: string;
    }>;
    fetchNFTs(address: string, chain: EVMBlockchainIncludingTestnet): Promise<any>;
    getPasskeyServerUrl(): string;
}

interface EVMToken {
    chain: EVMBlockchainIncludingTestnet;
    contractAddress: string;
}
interface NFTEVMToken extends EVMToken {
    tokenId: string;
    type: "nft";
}
interface SFTEVMToken extends EVMToken {
    tokenId: string;
    type: "sft";
}
interface ERC2OEVMToken extends EVMToken {
    type: "ft";
}
type ERC20TransferType = {
    token: ERC2OEVMToken;
    amount: bigint;
};
type SFTTransferType = {
    token: SFTEVMToken;
    quantity: number;
};
type NFTTransferType = {
    token: NFTEVMToken;
};
type TransferType = ERC20TransferType | SFTTransferType | NFTTransferType;

/**
 * Smart wallet interface for EVM chains enhanced with Crossmint capabilities.
 * Core functionality is exposed via [viem](https://viem.sh/) clients within the `client` property of the class.
 */
declare class EVMSmartWallet extends LoggerWrapper {
    private readonly crossmintService;
    private readonly accountClient;
    readonly chain: EVMBlockchainIncludingTestnet;
    /**
     * [viem](https://viem.sh/) clients that provide an interface for core wallet functionality.
     */
    readonly client: {
        /**
         * An interface to interact with the smart wallet, execute transactions, sign messages, etc.
         */
        wallet: SmartWalletClient;
        /**
         * An interface to read onchain data, fetch transactions, retrieve account balances, etc. Corresponds to public [JSON-RPC API](https://ethereum.org/en/developers/docs/apis/json-rpc/) methods.
         */
        public: PublicClient;
    };
    constructor(crossmintService: CrossmintWalletService, accountClient: SmartWalletClient, publicClient: PublicClient<HttpTransport>, chain: EVMBlockchainIncludingTestnet);
    /**
     * The address of the smart wallet.
     */
    get address(): `0x${string}`;
    /**
     * @returns The transaction hash.
     */
    transferToken(toAddress: string, config: TransferType): Promise<string>;
    /**
     * @returns A list of NFTs owned by the wallet.
     */
    nfts(): Promise<any>;
}

declare class SmartWalletSDK extends LoggerWrapper {
    private readonly smartWalletService;
    private readonly errorProcessor;
    private constructor();
    /**
     * Initializes the SDK with the **client side** API key obtained from the Crossmint console.
     * @throws error if the api key is not formatted correctly.
     */
    static init({ clientApiKey }: SmartWalletSDKInitParams): SmartWalletSDK;
    /**
     * Retrieves or creates a wallet for the specified user.
     * The default configuration is a `PasskeySigner` with the name, which is displayed to the user during creation or signing prompts, derived from the provided jwt.
     *
     * Example using the default passkey signer:
     * ```ts
     * const wallet = await smartWalletSDK.getOrCreateWallet({ jwt: "xxx" }, "base");
     * ```
     */
    getOrCreateWallet(user: UserParams, chain: EVMBlockchainIncludingTestnet, walletParams?: WalletParams): Promise<EVMSmartWallet>;
}

export { AdminAlreadyUsedError, AdminMismatchError, ConfigError, CrossmintServiceError, type EOASigner, type ERC20TransferType, EVMSmartWallet, JWTDecryptionError, JWTExpiredError, JWTIdentifierError, JWTInvalidError, type NFTTransferType, NonCustodialWalletsNotEnabledError, NotAuthorizedError, OutOfCreditsError, PasskeyMismatchError, type PasskeySigner, type SFTTransferType, SmartWalletSDK, SmartWalletSDKError, type SmartWalletSDKInitParams, TransferError, type TransferType, type UserParams, UserWalletAlreadyCreatedError, type ViemAccount, type WalletParams };
